
/************************************************************************
* Copyright (c) 1995 Valentin Shopov, Blagoevgrad, Bulgaria		*
*									*
* Permission to use, copy, modify, and distribute this material for any	*
* purpose and without fee is hereby granted, provided that the above	*
* copyright notice and this permission notice appear in all copies, and	*
* that the name of the author not be used in advertising or publicity 	*
* pertaining to this material without the specific, prior written 	*
* permission of the author.						*
*									*
* THE AUTHOR MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY	*  
* OF THIS MATERIAL FOR ANY PURPOSE. IT IS PROVIDED "AS IS", WITHOUT ANY	* 
* EXPRESS OR IMPLIED WARRANTIES.					*
*************************************************************************/

/**************************ciilib2.0.2***********************************
 *									*
 *        Character to Inline Image Conversion Library (CIILIB)		*
 *									*
 *	Encoder for Character by Inline Image of HTML on HTTP		*
 *			by Valentin Shopov				*
 *			valentin@rs.aubg.bg				*
 ************************************************************************/

#include <stdio.h>
#include <memory.h>
#include "cii.h"

/*////////////////////////////////////////////////////////////////////// */


/*
 *    CII output image type selector
 */
CII_setitype (sitype)
     char *sitype;
{
  ;
}

/*
 *    text to CII encoder
 */
CII_toCII (any_text, cii_text, ctype, charsets, urlbase)
     unsigned char *any_text, *cii_text;
     char *ctype, *charsets[];
     char *urlbase;
{
  static _CII_toCII ();
  return _CII_toCII (any_text, cii_text, ctype, charsets, urlbase);
}


static int isTag, isTitle, isAnchor, isMozilla, TagBufLen;
static int isKOI, isJP, isBIG5;
static char *Hchar = "k";
static unsigned char TagBuf[2048];

extern struct CharCodesXBM CII_CC[];

static struct CIIoptions CII_O =
{"", 60, 0, 0, 0, 0, 0};


static int txColor = 0x000000;
static int lkColor = 0x0000ee;
static char *Pchar = "N";

static int
CharsetIDh (abbr, height)
     char *abbr;
{
  int i;
  char *nick;

  for (i = 0; nick = CII_CC[i].name; i++)
    {
      if (height == 0 || CII_CC[i].height == height)
	{
	  if (strncasecmp (nick, abbr, strlen (abbr)) == 0)
	    return i;
	  if (strcasecmp (abbr, CII_CC[i].regname) == 0)
	    return i;
	}
    }
  return -1;
}

static int
CharsetID (abbr)
     char *abbr;
{
  return CharsetIDh (abbr, 0);
}

static int
to256c (csname)
     char *csname;
{
  int cc_id;

  if ((cc_id = CharsetID (csname)) == -1)
    return 0;
  return CII_CC[cc_id].is2bytes == 0;
}


static unsigned char
koi8tocyr (kchar)
     unsigned char kchar;
{
  char *cyrch = "\356\320\321\346\324\325\344\323\345\330\331\332\333\334\
\335\336\337\357\340\341\342\343\326\322\354\353\327\350\355\351\347\352";
  if (kchar >= 0xc0 && kchar <= 0xdf)
    return cyrch[kchar - 0xc0];
  else if (kchar >= 0xe0 && kchar <= 0xff)
    return cyrch[kchar - 0xe0] - 32;
  else
    return kchar;
}



static int
strncscmp (str1, str2, snum)
     char *str1, *str2;
     int snum;
{
  char ch1;
  int i, j, isSPC;
  for (i = 0, j = 0, isSPC = 0; i < snum && str1[j]; i++)
    {
      for (; str1[j] <= ' ' && str1[j];)
	{
	  j++;
	  isSPC = 1;
	}
      ch1 = str1[j];
      if (ch1 >= 'a' && ch1 <= 'z')
	ch1 -= 32;
      if (isSPC)
	{
	  isSPC = 0;
	  if (ch1 == '=' && str2[i] == '=')
	    {
	      j++;
	      continue;
	    }
	  if (str2[i - 1] == '=')
	    {
	      i--;
	      continue;
	    }
	  if (ch1 == '>' && str2[i] == '>')
	    {
	      j++;
	      continue;
	    }

	  if (str2[i] != ' ')
	    return 1;
	}
      else
	{
	  if (str2[i] != ch1)
	    return 1;
	  j++;
	}
    }
  return 0;
}

static int
isN256C (src)
     unsigned char *src;
{
  register unsigned char *sp;
  int is8bit = 0;

  for (sp = src; *sp; sp++)
    {
      if (*sp >= 0x80)
	{
	  is8bit++;
	  if (*sp < 0xa0)
	    return 1;
	}
    }
  return (is8bit) ? 0 : -1;
}

static int
rvrsBUF (src)
     unsigned char *src;
{
  register unsigned char *sp, *dp;
  unsigned char *tmpb;
  tmpb = (unsigned char *) strdup (src);
  for (sp = (unsigned char *) strrchr (tmpb, 0) - 1, dp = src; *dp; sp--, dp++)
    *dp = *sp;
  free (tmpb);
}

GIFCII_setccid (Hsize, charset)
     int Hsize;
{
  int cc_id;

  switch (Hsize)
    {
    case 'h':
      if ((cc_id = CharsetIDh (charset, 24)) == -1)
	cc_id = CharsetIDh (charset, 14);
      break;
    case 'i':
      if ((cc_id = CharsetIDh (charset, 16)) == -1)
	cc_id = CharsetIDh (charset, 14);
      break;
    case 'j':
    case 'k':
      if ((cc_id = CharsetIDh (charset, 14)) == -1)
	cc_id = CharsetIDh (charset, 16);
      break;
    default:
      return -1;
    }

  if (cc_id == -1)
    cc_id = CharsetID (charset);

  return cc_id;
}

static
GIFCII_setfsize (chHp, chWp, Acharp, charset)
     int *chHp, *chWp;
     char **Acharp;
     char *charset;
{
  int cc_id;

  cc_id = GIFCII_setccid (*Hchar, charset);
  *chHp = CII_CC[cc_id].height;
  *chWp = CII_CC[cc_id].width;

  *Acharp = "";
  if (CII_O.uline)
    if (isAnchor && isMozilla)
      {
	*Acharp = "u";
	*chHp += 2;
      }
}

static int
GIFCII_switch_H (sp)
     char *sp;
{
  static int isHdr[6];
  static char HdrIn[] = "<H1>";
  static char HdrOut[] = "</H1>";
  static char *HcharAr[6] =
  {"h", "i", "j", "j", "j", "j"};
  int i;

  for (i = 0; i < 6; i++)
    {
      HdrIn[2] = HdrOut[3] = '1' + i;
      if (!isHdr[i] && !strncscmp (sp, HdrIn, sizeof (HdrIn) - 1))
	{
	  isHdr[i] = 1;
	  Hchar = HcharAr[i];
	  return 1;
	}
      if (isHdr[i] && !strncscmp (sp, HdrOut, sizeof (HdrOut) - 1))
	{
	  isHdr[i] = 0;
	  Hchar = "k";
	  return 1;
	}
    }

  return 0;
}

static int
getColor (colorp, sp)
     int *colorp;
     char *sp;
{
  int i = 6, j, color = 0;

  while (sp[i] == ' ' || sp[i] == '\t')
    i++;
  if (sp[i] == '\"')
    i++;
  if (sp[i] == '#')
    i++;

  for (j = i; j < i + 6; j++)
    {
      if (sp[j] == 0 || sp[j] == ' ' || sp[j] == '\"' || sp[j] == '\n')
	break;

      color *= 0x10;

      if (sp[j] >= '0' && sp[j] <= '9')
	color += sp[j] - '0';
      else if (sp[j] >= 'A' && sp[j] <= 'F')
	color += sp[j] - 'A' + 10;
      else if (sp[j] >= 'a' && sp[j] <= 'f')
	color += sp[j] - 'a' + 10;
    }

  *colorp = color;
  return j - 1;
}
static int
GIFCII_switch_Color (sp, isMozilla)
     char *sp;
     int isMozilla;
{
  int i;
  if (isMozilla && !strncscmp (sp, "<BODY", 5))
    {
      int isText = 0, isLink = 0, j;
      for (i = 0; sp[i] != '>' && sp[i]; i++)
	{
	  if (sp[i] == ' ' || sp[i] == '\"' || sp[i] == '\n')
	    {
	      if (!strncscmp (sp + i + 1, "BACKGROUND=", 11))
		Pchar = "I";
	      if (!isText && !strncscmp (sp + i + 1, "TEXT=", 5))
		{
		  isText = 1;
		  i += getColor (&txColor, sp + i);
		}
	      if (!isLink && !strncscmp (sp + i + 1, "LINK=", 5))
		{
		  isLink = 1;
		  i += getColor (&lkColor, sp + i);
		}
	    }
	}
      return 1;
    }
  else
    return 0;
}

static unsigned char *
GIFCII_putCII (dp, sp, ubase, charset, isMozilla, do256, isLAP)
     unsigned char *dp, *sp, *ubase, *charset;
{
  unsigned char c0, c1;
  int w, lastloop = 0, firstloop = 1, chH, chW;
  char *Achar;
  char *Gchar;
  Gchar = (CII_O.b2xbm && !(isAnchor ? lkColor : txColor)) ? "x" : "";


  if (!*sp)
    return sp;

  GIFCII_setfsize (&chH, &chW, &Achar, charset);

  if (*Hchar == 'h' && chH < 24)
    {
      chW <<= 1;
      chH <<= 1;
    }

  for (w = 0;; w++)
    {
      if (!(sp[0] && (sp[0] & 0x80) && (do256 || (sp[1] & 0x80) || isBIG5)))
	lastloop = 1;

      if (!firstloop && (w == CII_O.gwidth && !isLAP || w == 0 || lastloop))
	{
	  if (isMozilla)
	    {
	      sprintf (dp, "\" WIDTH=%d HEIGHT=%d%s>",
		       chW * ((w == 0 || isLAP) ? 1 : w),
		       chH, isAnchor ? " BORDER=0" : "");
	    }
	  else
	    sprintf (dp, "\">");
	  dp += strlen (dp);
	  w = 0;
	  if (lastloop)
	    break;
	}

      if (w == 0)
	{
	  sprintf (dp, "<IMG SRC=\"%s%s/%06x%s%s%s%s", ubase, charset,
		(isAnchor ? lkColor : txColor), Pchar, Gchar, Hchar, Achar);
	  dp += strlen (dp);
	  if (isLAP)
	    *dp++ = '+';
	}

      if (do256)
	{
	  c0 = 0;
	  c1 = (isKOI == 1) ? koi8tocyr (*sp) : *sp;
	  sp++;
	}
      else
	{
	  c0 = *sp++;
	  c1 = *sp++;
	}
      sprintf (dp, "%02x%02x", c0, c1);
      dp += 4;

      if (CII_O.schar)
	if (!isLAP && do256 && !isAnchor && *Hchar == 'k')
	  w = -1;

      if (CII_O.skana)
	if (isJP && !isAnchor && *Hchar == 'k')
	  if ((c0 >= 0xa1 && c0 <= 0xa8)
	      || (sp[0] >= 0xa1 && sp[0] <= 0xa8 && (sp[1] & 0x80)))
	    w = -1;

      if (CII_O.uspc)
	if (isAnchor && *sp == ' ')
	  {
	    for (; (w + 1) < CII_O.gwidth && *sp == ' '; w++)
	      {
		sprintf (dp, "USPC");
		sp++;
		dp += 4;
	      }
	  }
      firstloop = 0;
    }
  return sp;
}

extern char *HTTP_getRequestField ();
unsigned char *
EUC_TO_CII (src, dst, ctype, charsetC, ubase, isLAP)
     unsigned char *src, *dst;
     char *ctype, *charsetC, *ubase;
{
  register unsigned char *sp, *dp;
  char *charset, *Achar, *ltPTR, *gtPTR;
  unsigned char t256c;
  int isPlain, do256, AncLen, i, chW, chH;

  isPlain = strcasecmp (ctype, "text/plain") == 0;


  i = CharsetID (isKOI == 1 ? "cyr" : charsetC);

  if (i == -1)
    {
      strcpy (dst, src);
      return dst;
    }
  charset = CII_CC[i].regname;

  isJP = (CharsetID (charset) == CharsetID ("jp"));
  isBIG5 = (CharsetID (charset) == CharsetID ("big5"));
  do256 = to256c (charset);

  ltPTR = (char *) strrchr (src, '<');
  gtPTR = (char *) strrchr (src, '>');

  if ((gtPTR && ltPTR - gtPTR > 0) || ((TagBufLen || ltPTR) && !gtPTR))
    {
      for (i = 0; src[i]; i++)
	{
	  TagBuf[TagBufLen++] = src[i];
	}
      TagBuf[TagBufLen] = '\0';
      *dst = '\0';
      return dst;
    }

  if (TagBufLen)
    {
      for (i = 0; src[i]; i++)
	{
	  TagBuf[TagBufLen++] = src[i];
	}
      TagBuf[TagBufLen] = '\0';
    }



  for (sp = (TagBufLen) ? TagBuf : src, dp = dst; *sp;)
    {
      if (*sp == '<' && !isTag && !isPlain)
	{

	  isTag = 1;
	  if (GIFCII_switch_H (sp))
	    ;
	  else if (GIFCII_switch_Color (sp, isMozilla))
	    ;
	  else if (!isAnchor && !strncscmp (sp, "<A HREF=", 8))
	    {
	      isAnchor = 1;
	    }
	  else if (isAnchor && !strncscmp (sp, "</A>", 4))
	    {
	      isAnchor = 0;
	      isTag = 0;
	    }
	  else if (!isTitle && !strncscmp (sp, "<TITLE>", 7))
	    isTitle = 1;
	  else if (isTitle && !strncscmp (sp, "</TITLE>", 8))
	    isTitle = 0;
	}
      if (isTag || isTitle)
	{
	  if (*sp == '>' && isTag)
	    isTag = 0;
	  *dp++ = *sp++;
	}
      else if (*sp & 0x80)
	{
	  if (do256 || (sp[1] & 0x80) || isBIG5)
	    {
	      sp = GIFCII_putCII (dp, sp, ubase, charset, isMozilla, do256, isLAP);
	      dp += strlen (dp);
	    }
	  else
	    sp++;
	}
      else if (isPlain)
	{
	  switch (*sp)
	    {
	    case '<':
	      sprintf (dp, "&lt;");
	      dp += 4;
	      sp++;
	      break;
	    case '>':
	      sprintf (dp, "&gt;");
	      dp += 4;
	      sp++;
	      break;
	    case '&':
	      sprintf (dp, "&amp;");
	      dp += 5;
	      sp++;
	      break;
	    case '"':
	      sprintf (dp, "&quot;");
	      dp += 6;
	      sp++;
	      break;
	    default:
	      *dp++ = *sp++;
	      break;
	    }
	}
      else
	*dp++ = *sp++;
    }
  *dp = 0;
  if (TagBufLen)
    TagBufLen = 0;

  return dst;
}

#define ESC     '\033'
#define TO96    ','
#define TO2B    '$'
#define TO94    '('
#define TOASC1  'B'
#define TOASC2  'J'
#define TOASC3  'H'
#define TOGB    'A'
#define TOJP    'B'
#define TOJPOLD '@'
#define TOKSP   'D'
#define TORK    'I'
#define TOKR    'C'
#define TOBIG51 '0'
#define TOBIG52 '1'
#define TOLTN1  'A'
#define TOLTN2  'B'
#define TOLTN3  'C'
#define TOLTN4  'D'
#define TOCYR   'L'
#define TOARA   'G'
#define TOGRE   'F'
#define TOHEB   'H'
#define TOLTN5  'M'
#define TOVIET  '1'
#define TOTHAI  'T'
#define TOARAB1	'3'
#define TOARAB2	'4'
#define TOETHIO '2'
#define R2L1    '['
#define R2L3    ']'
#define R2LIN   '2'
#define R2LOUT  '0'
#define TOG0    '0'
#define TOG1    '1'

ISO2022_toCII (src, dst, ctype, charsets, ubase)
     unsigned char *src, *dst;
     char *ctype, *charsets[], *ubase;
{
  unsigned char *tmpbuf, *csnbuf, *tmpdst, *euc;
  register unsigned char *sp, *dp, *cp;
  static char *Charsets[] =
  {
    "jp", "jp", "ksp", "kr", "gb", "big5", "ethio",
    "ltn1", "ltn2", "ltn3", "ltn4", "cyr", "ara",
    "gre", "heb", "ltn5", "viet", "thai",
    "rk", "arab1", "arab2", 0};
  enum
    {
      csn_0, csn_jp, csn_ksp, csn_kr, csn_gb, csn_big5, csn_ethio,
      csn_ltn1, csn_ltn2, csn_ltn3, csn_ltn4, csn_cyr, csn_ara,
      csn_gre, csn_heb, csn_ltn5, csn_viet, csn_thai,
      csn_rk, csn_arab1, csn_arab2
    };

  int csn = 0, csnprev, isR2L = 0, isR2Lprev, isAscii = 1;
  int isG0 = 0, isG0prev, toCII, lcBIG5, i;

  tmpdst = dst;
  dp = tmpbuf = (unsigned char *) malloc (strlen (src) + 1);
  cp = csnbuf = (unsigned char *) malloc (strlen (src) + 1);

  for (sp = src; *sp;)
    {
      toCII = 0;
      csnprev = csn;
      isR2Lprev = isR2L;
      isG0prev = isG0;

      if (sp[0] == ESC)
	{
	  if (sp[1] == TO2B)
	    {
	      i = sp[2] == TO94 ? 3 : 2;
	      isAscii = 0;
	      switch (sp[i])
		{
		case TOJP:
		case TOJPOLD:
		  csn = csn_jp;
		  break;
		case TOKSP:
		  csn = csn_ksp;
		  break;
		case TOKR:
		  csn = csn_kr;
		  break;
		case TOGB:
		  csn = csn_gb;
		  break;
		case TOBIG51:
		  csn = csn_big5;
		  lcBIG5 = 0;
		  break;
		case TOBIG52:
		  csn = csn_big5;
		  lcBIG5 = 1;
		  break;
		case TOETHIO:
		  csn = csn_ethio;
		  break;
		default:
		  isAscii = 1;
		  *dp++ = *sp++;
		}
	      if (!isAscii)
		{
		  sp += i + 1;
		  if ((csn != csnprev || csn == csn_big5) && !isR2L)
		    toCII = 1;
		}
	    }
	  else if (sp[1] == TO94)
	    {
	      isAscii = 0;
	      switch (sp[2])
		{
		case TORK:
		  csn = csn_rk;
		  break;
		case TOARAB1:
		  csn = csn_arab1;
		  break;
		case TOARAB2:
		  csn = csn_arab2;
		  break;
		case TOASC1:
		case TOASC2:
		case TOASC3:
		  isAscii = 1;
		  sp += 3;
		  break;
		default:
		  isAscii = 1;
		  *dp++ = *sp++;
		}
	      if (!isAscii)
		{
		  sp += 3;
		  if (csn != csnprev && !isR2L)
		    toCII = 1;
		}
	    }
	  else if (sp[1] == TO96)
	    {
	      isAscii = 0;
	      switch (sp[2])
		{
		case TOLTN1:
		  csn = csn_ltn1;
		  break;
		case TOLTN2:
		  csn = csn_ltn2;
		  break;
		case TOLTN3:
		  csn = csn_ltn3;
		  break;
		case TOLTN4:
		  csn = csn_ltn4;
		  break;
		case TOCYR:
		  csn = csn_cyr;
		  break;
/*
   case TOARA:
   csn = csn_ara;
   break;
 */
		case TOGRE:
		  csn = csn_gre;
		  break;
		case TOHEB:
		  csn = csn_heb;
		  break;
		case TOLTN5:
		  csn = csn_ltn5;
		  break;
		case TOVIET:
		  csn = csn_viet;
		  break;
		case TOTHAI:
		  csn = csn_thai;
		  break;
		default:
		  isAscii = 1;
		  *dp++ = *sp++;
		}
	      if (!isAscii)
		{
		  sp += 3;
		  if (csn != csnprev && !isR2L)
		    toCII = 1;
		}
	    }
	  else if (sp[1] == TOG0 || sp[1] == TOG1)
	    {
	      isG0 = (sp[1] == TOG0);
	      sp += 2;
	      toCII = 1;
	    }
	  else if (sp[1] == R2L1 && sp[3] == R2L3)
	    {
	      if (sp[2] == R2LIN)
		{
		  sp += 4;
		  if (!isR2L)
		    {
		      isR2L = 1;
		      toCII = 1;
		    }
		}
	      else if (sp[2] == R2LOUT)
		{
		  sp += 4;
		  if (isR2L)
		    {
		      isR2L = 0;
		      toCII = 1;
		    }
		}
	      else
		*dp++ = *sp++;
	    }
	  else
	    *dp++ = *sp++;
	}
      else if (csn == csn_big5 && !isAscii && sp[1])
	{
	  unsigned int temp_i = (sp[0] - 0x21) * (0xFF - 0xA1) + (sp[1] - 0x21);
	  if (lcBIG5 == 1)
	    temp_i += (0xFF - 0xA1 + 0x7F - 0x40) * (0xC9 - 0xA1);
	  dp[0] = temp_i / (0xFF - 0xA1 + 0x7F - 0x40) + 0xA1;
	  dp[1] = temp_i % (0xFF - 0xA1 + 0x7F - 0x40);
	  dp[1] += dp[1] < 0x3F ? 0x40 : 0x62;
	  sp += 2;
	  dp += 2;
	}
      else
	{
	  if (isR2L)
	    *cp++ = csn;
	  *dp++ = (csn && !isAscii && *sp > 0x20 && *sp <= 0x7f) ? *sp | 0x80 : *sp;
	  sp++;
	}

      if (csn == csn_thai && isAscii)
	{
	  toCII = 1;
	  csn = 0;
	}

      if (toCII || !*sp)
	{
	  if (isR2Lprev)
	    {
	      int csnR2L;
	      unsigned char *buf = (unsigned char *) malloc (strlen (tmpbuf) + 1);
	      unsigned char *bp = buf;
	      while (dp > tmpbuf)
		{
		  *bp++ = *--dp;
		  csnR2L = *--cp;
		  if (csnR2L != *(cp - 1) || dp == tmpbuf)
		    {
		      *bp = 0;
		      EUC_TO_CII (buf, tmpdst, ctype, Charsets[csnR2L], ubase, 0);
		      tmpdst += strlen (tmpdst);
		      bp = buf;
		    }
		}
	      free (buf);
	      cp = csnbuf;
	    }
	  else
	    {
	      *dp = '\0';
	      if (!csnprev)
		{
		  euc = (unsigned char *) strdup (tmpbuf);
		  TO_EUC (euc, tmpbuf);
		  free (euc);
		}
	      EUC_TO_CII (tmpbuf, tmpdst, ctype, Charsets[csnprev], ubase, isG0prev);
	      tmpdst += strlen (tmpdst);
	      dp = tmpbuf;
	    }
	}
    }
  free (tmpbuf);
  free (csnbuf);
}

HZ_TO_CII (src, dst, ctype, ubase)
     unsigned char *src, *dst;
     char *ctype, *ubase;
{
  unsigned char *tmpbuf, *tmpdst;
  register unsigned char *sp, *dp;
  int inHZ = 0, toCII, c, len;

  tmpdst = dst;
  tmpbuf = (unsigned char *) malloc (strlen (src) + 1);
  for (sp = src, dp = tmpbuf; *sp;)
    {
      toCII = 0;
      len = 1;
      c = *sp;

      if (c == '~')
	{
	  if (inHZ)
	    {
	      if (sp[1] == '}')
		{
		  inHZ = 0;
		  len = 0;
		  sp += 2;
		}
	    }
	  else
	    {
	      if (sp[1] == '{')
		{
		  inHZ = 1;
		  len = 0;
		  sp += 2;
		  toCII = 1;
		}
	      else if (sp[1] == '~')
		{
		  sp++;
		}
	    }
	}
      else if (c == '&')
	{
	  if (!strncmp (sp, "&lt;", 4))
	    {
	      len = 4;
	      c = '<';
	    }
	  else if (!strncmp (sp, "&gt;", 4))
	    {
	      len = 4;
	      c = '>';
	    }
	  else if (!strncmp (sp, "&amp;", 5))
	    {
	      len = 5;		/* c = '&'; */
	    }
	  else if (!strncmp (sp, "&quot;", 6))
	    {
	      len = 6;
	      c = '\"';
	    }
	}

      if (len)
	{
	  if (inHZ && c > 0x20 && c <= 0x7f)
	    {
	      c |= 0x80;
	    }
	  *dp++ = c;
	  sp += len;
	}

      if (toCII || !sp[1])
	{
	  *dp = '\0';
	  EUC_TO_CII (tmpbuf, tmpdst, ctype, "gb", ubase, 0);
	  tmpdst += strlen (tmpdst);
	  dp = tmpbuf;
	}
    }
  free (tmpbuf);
}

static char *
chcode (str)
     unsigned char *str;
{
  unsigned char *s, ch;
  int n8, n8s, nhz;

  n8 = 0;
  n8s = 0;
  nhz = 0;

  for (s = str; ch = *s; s++)
    {
      if (ch == '~')
	if (s[1] == '{' || s[1] == '}')
	  nhz++;

      if (ch & 0x80)
	{
	  n8++;
	  n8s++;
	}
      else
	{
	  if (n8 % 2 != 0)
	    return "big5";
	  n8 = 0;
	}
    }
  if (2 <= nhz)
    return "hz";
  if (0 < n8s)
    return "gb";
  return "jp";
}



extern char DELEGATE_FLAGS[];
static char URLcharcode[64];

static void
CIIinit (charset)
     char *charset;
{

  char URLOptions[128], *IFptr, *DFptr, *URLopt;

  CII_O.gwidth = GIFWIDTH;
#ifdef UNDERLINE
  CII_O.uline = 1;
#endif
#ifdef USPACE
  CII_O.uspc = 1;
#endif
#ifdef BLACKTXTASXBM
  CII_O.b2xbm = 1;
#endif
#ifdef LONGKANJIONLY
  CII_O.skana = 1;
#endif
#ifdef NOLONG256C
  CII_O.schar = 1;
#endif


  DFptr = DELEGATE_FLAGS;
  URLopt = URLOptions;
  *URLopt = 0;
  if ((IFptr = strstr (DFptr, "I.")) != 0)
    {
      IFptr += 2;
      for (; *IFptr != '.' && *IFptr; IFptr++, URLopt++)
	{
	  *URLopt = *IFptr;
	  if (*URLopt >= 'a' && *URLopt <= 'z')
	    *URLopt -= 32;
	}
      *URLopt = 0;
    }

  URLopt = URLcharcode;
  *URLopt = 0;

  if ((IFptr = strstr (URLOptions, "CC:")) != 0)
    {
      IFptr += 3;
      for (; *IFptr != '/' && *IFptr; IFptr++, URLopt++)
	*URLopt = *IFptr;
      *URLopt = 0;
    }

  CII_O.charcode = (URLcharcode[0]) ? URLcharcode : charset;

  if ((IFptr = strstr (URLOptions, "GW:")) != 0)
    {
      IFptr += 3;
      CII_O.gwidth = atoi (IFptr);
      if (CII_O.gwidth < 1 || CII_O.gwidth > GIFWIDTH)
	CII_O.gwidth = GIFWIDTH;
    }

  if ((IFptr = strstr (URLOptions, "UL:")) != 0)
    {
      IFptr += 3;
      CII_O.uline = (*IFptr == 'Y') ? 1 : ((*IFptr == 'N') ? 0 : CII_O.uline);
    }
  if ((IFptr = strstr (URLOptions, "US:")) != 0)
    {
      IFptr += 3;
      CII_O.uspc = (*IFptr == 'Y') ? 1 : ((*IFptr == 'N') ? 0 : CII_O.uspc);
    }
  if ((IFptr = strstr (URLOptions, "BX:")) != 0)
    {
      IFptr += 3;
      CII_O.b2xbm = (*IFptr == 'Y') ? 1 : ((*IFptr == 'N') ? 0 : CII_O.b2xbm);
    }
  if ((IFptr = strstr (URLOptions, "SK:")) != 0)
    {
      IFptr += 3;
      CII_O.skana = (*IFptr == 'Y') ? 1 : ((*IFptr == 'N') ? 0 : CII_O.skana);
    }
  if ((IFptr = strstr (URLOptions, "SC:")) != 0)
    {
      IFptr += 3;
      CII_O.schar = (*IFptr == 'Y') ? 1 : ((*IFptr == 'N') ? 0 : CII_O.schar);
    }


}


static int initDone = 0;

static
_CII_toCII (any, gr, ctype, charsets, ubase)
     unsigned char *any, *gr;
     char *ctype, *charsets[];
     char *ubase;
{
  unsigned char *euc;
  int i, strC;
  char *charsetsII;
  char UserAGENT[1024], *mozi;

  if (!initDone)
    {
      CIIinit (charsets[0]);
      initDone = 1;
    }


  HTTP_getRequestField ("User-Agent", UserAGENT, sizeof (UserAGENT));
  if (strstr (UserAGENT, "Emacs-W3") || strstr (UserAGENT, "Lynx"))
    {
      strcpy (gr, any);
      return;
    }

  if ((mozi = (char *) strstr (UserAGENT, "Mozilla")) && strcmp (mozi, "Mozilla/1.1") >= 0)
    isMozilla = 1;
  else
    {
      isMozilla = 0;
      lkColor = 0x0000b0;
    }

  strC = isN256C (any);
  if (isKOI != 2 && CharsetID ("koi8") != -1)
    isKOI = 2;
  else if (!strncasecmp (CII_O.charcode, "koi8", 4))
    isKOI = 1;
  else
    isKOI = 0;

  if (strC >= 0 && (isKOI == 1 || to256c (CII_O.charcode)))
    EUC_TO_CII (any, gr, ctype, CII_O.charcode, ubase, 0);
  else if (strC < 0 && strchr (any, 0x1b))
    {
      if (isKOI == 1)
	isKOI = 0;
      ISO2022_toCII (any, gr, ctype, charsets, ubase);
    }
  else if (strcasecmp (CII_O.charcode, "cn") == 0)
    {
      charsetsII = chcode (any);
      if (charsetsII[0] == 'h')
	HZ_TO_CII (any, gr, ctype, ubase);
      else
	EUC_TO_CII (any, gr, ctype, charsetsII, ubase, 0);
    }
  else if (CharsetID (CII_O.charcode) == CharsetID ("jp"))
    {
      euc = (unsigned char *) malloc (strlen (any) + 1);
      TO_EUC (any, euc);
      EUC_TO_CII (euc, gr, ctype, CII_O.charcode, ubase, 0);
      free (euc);
    }
  else
    EUC_TO_CII (any, gr, ctype, CII_O.charcode, ubase, 0);
}



static int
CII_INFO (tc, vno, iname)
     FILE *tc;
     char *iname;
{
  char infobuf[5000];
  char useragentbuf[1024];
  static char *y = "Yes", *n = "No";
  int leng = 0;
  /* HTTP_getRequestField ("CII-Version", ciiverbuf, sizeof (ciiverbuf)); */
  HTTP_getRequestField ("User-Agent", useragentbuf, sizeof (useragentbuf));



  CIIinit ("");


  sprintf (infobuf, "\
<TITLE>CIILIB Information</TITLE>\n\
<p><h1><center>Welcome to CIILIB %s!!!</center></h1>\n\
<hr><h2><pre>\n\
GIFwidth:   %d\n\
Underline:  %s\n\
Uspace:     %s\n\
Black2XBM:  %s\n\
ShortKana:  %s\n\
ShortChars: %s\n\
User-Agent: %s\n\
</pre><hr><p>Thanks !</h2>\n\
<br><address>\n\
<a href=\"mailto:valentin@rs.aubg.bg\">\n\
valentin@rs.aubg.bg(Valentin Shopov)\n\
</a>\n\
</address>\n\
 ", CIIVER, CII_O.gwidth, ((CII_O.uline) ? y : n), ((CII_O.uspc) ? y : n), ((CII_O.b2xbm) ? y : n), ((CII_O.skana) ? y : n), ((CII_O.schar) ? y : n), useragentbuf);

  leng += HTTP_putHeader (tc, vno, "text/html", 0);
  fputs (infobuf, tc);
  leng += strlen (infobuf);
  return leng;

}

/*
 *    CII font server
 */
CII_putHTTP (out, vno, cii_code)
     FILE *out;
     char *cii_code;
{
  char *inf;
  if (!strcasecmp (cii_code, "info"))
    return CII_INFO (out, vno, cii_code);
  else
    return CII_putHTTP_GIF (out, vno, cii_code);
}
